-module(erl_parse).
-define(THIS_MODULE, erl_parse).
-export([parse/1, parse_and_scan/1, format_error/1]).

%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%

-export([parse_form/1,parse_exprs/1,parse_term/1]).
-export([normalise/1,abstract/1,tokens/1,tokens/2]).
-export([abstract/2, package_segments/1]).
-export([inop_prec/1,preop_prec/1,func_prec/0,max_prec/0]).

%% The following directive is needed for (significantly) faster compilation
%% of the generated .erl file by the HiPE compiler.  Please do not remove.
-compile([inline,{hipe,[{regalloc,linear_scan}]}]).


%% mkop(Op, Arg) -> {op,Line,Op,Arg}.
%% mkop(Left, Op, Right) -> {op,Line,Op,Left,Right}.

mkop(L, {Op,Pos}, R) -> {op,Pos,Op,L,R}.

mkop({Op,Pos}, A) -> {op,Pos,Op,A}.

%% keep track of line info in tokens
line(Tup) -> element(2, Tup).

%% Entry points compatible to old erl_parse.
%% These really suck and are only here until Calle gets multiple
%% entry points working.

parse_form(Tokens) ->
    parse(Tokens).

parse_exprs(Tokens) ->
    case parse([{atom,0,f},{'(',0},{')',0},{'->',0}|Tokens]) of
	{ok,{function,Lf,f,0,[{clause,Lc,[],[],Exprs}]}} ->
	    {ok,Exprs};
	{error,E} -> {error,E}
    end.

parse_term(Tokens) ->
    case parse([{atom,0,f},{'(',0},{')',0},{'->',0}|Tokens]) of
	{ok,{function,Lf,f,0,[{clause,Lc,[],[],[Expr]}]}} ->
	    case catch normalise(Expr) of
		{'EXIT',R} ->
		    {error,{line(Expr),?MODULE,"bad term"}};
		Term -> {ok,Term}
	    end;
	{ok,{function,Lf,f,0,[{clause,Lc,[],[],[E1,E2|Es]}]}} ->
	    {error,{line(E2),?MODULE,"bad term"}};
	{error,E} -> {error,E}
    end.

%% build_attribute(AttrName, AttrValue) ->
%%	{attribute,Line,module,Module}
%%	{attribute,Line,export,Exports}
%%	{attribute,Line,import,Imports}
%%	{attribute,Line,record,{Name,Inits}}
%%	{attribute,Line,file,{Name,Line}}
%%	{attribute,Line,Name,Val}

build_attribute({atom,La,module}, Val) ->
    case Val of
	[{atom,Lm,Module}] ->
	    {attribute,La,module,Module};
	[Name] ->
	    case package_segments(Name) of
		error ->
		    error_bad_decl(La, module);
		Module ->
		    {attribute,La,module,Module}
	    end;
	Other ->
	    error_bad_decl(La, module)
    end;
build_attribute({atom,La,export}, Val) ->
    case Val of
	[ExpList] ->
	    {attribute,La,export,farity_list(ExpList)};
	Other -> error_bad_decl(La, export)
    end;
build_attribute({atom,La,import}, Val) ->
    case Val of
	[Name] ->
	    case package_segments(Name) of
		error ->
		    error_bad_decl(La, import);
		Module ->
		    {attribute,La,import,Module}
	    end;
	[{atom,Lm,Mod},ImpList] ->
	    {attribute,La,import,{Mod,farity_list(ImpList)}};
	[Name, ImpList] ->
	    case package_segments(Name) of
		error ->
		    error_bad_decl(La, import);
		Module ->
		    {attribute,La,import,{Module,farity_list(ImpList)}}
	    end;
	Other -> error_bad_decl(La, import)
    end;
build_attribute({atom,La,record}, Val) ->
    case Val of
	[{atom,Ln,Record},RecTuple] ->
	    {attribute,La,record,{Record,record_tuple(RecTuple)}};
	Other -> error_bad_decl(La, record)
    end;
build_attribute({atom,La,file}, Val) ->
    case Val of
	[{string,Ln,Name},{integer,Ll,Line}] ->
	    {attribute,La,file,{Name,Line}};
	Other -> error_bad_decl(La, file)
    end;
build_attribute({atom,La,Attr}, Val) ->
    case Val of
	[Expr] ->
	    {attribute,La,Attr,term(Expr)};
	Other -> return_error(La, "bad attribute")
    end.

error_bad_decl(L, S) ->
    return_error(L, io_lib:format("bad ~w declaration", [S])).

farity_list({cons,Lc,{op,Lo,'/',{atom,La,A},{integer,Li,I}},Tail}) ->
    [{A,I}|farity_list(Tail)];
farity_list({nil,Ln}) -> [];
farity_list(Other) ->
    return_error(line(Other), "bad function arity").

record_tuple({tuple,Lt,Fields}) ->
    record_fields(Fields);
record_tuple(Other) ->
    return_error(line(Other), "bad record declaration").

record_fields([{atom,La,A}|Fields]) ->
    [{record_field,La,{atom,La,A}}|record_fields(Fields)];
record_fields([{match,Lm,{atom,La,A},Expr}|Fields]) ->
    [{record_field,La,{atom,La,A},Expr}|record_fields(Fields)];
record_fields([Other|Fields]) ->
    return_error(line(Other), "bad record field");
record_fields([]) -> [].

term(Expr) ->
    case catch normalise(Expr) of
	{'EXIT',R} -> return_error(line(Expr), "bad attribute");
	Term -> Term
    end.

package_segments(Name) ->
    package_segments(Name, [], []).

package_segments({record_field, _, F1, F2}, Fs, As) ->
    package_segments(F1, [F2 | Fs], As);
package_segments({atom, _, A}, [F | Fs], As) ->
    package_segments(F, Fs, [A | As]);
package_segments({atom, _, A}, [], As) ->
    lists:reverse([A | As]);
package_segments(_, _, _) ->
    error.

%% build_function([Clause]) -> {function,Line,Name,Arity,[Clause]}

build_function(Cs) ->
    Name = element(3, hd(Cs)),
    Arity = length(element(4, hd(Cs))),
    {function,line(hd(Cs)),Name,Arity,check_clauses(Cs, Name, Arity)}.

%% build_rule([Clause]) -> {rule,Line,Name,Arity,[Clause]'}

build_rule(Cs) ->
    Name = element(3, hd(Cs)),
    Arity = length(element(4, hd(Cs))),
    {rule,line(hd(Cs)),Name,Arity,check_clauses(Cs, Name, Arity)}.

%% build_fun(Line, [Clause]) -> {'fun',Line,{clauses,[Clause]}}.

build_fun(Line, Cs) ->
    Arity = length(element(4, hd(Cs))),
    {'fun',Line,{clauses,check_clauses(Cs, 'fun', Arity)}}.

check_clauses(Cs, Name, Arity) ->
     mapl(fun ({clause,L,N,As,G,B}) when N == Name, length(As) == Arity ->
		 {clause,L,As,G,B};
	     ({clause,L,N,As,G,B}) ->
		 return_error(L, "head mismatch") end, Cs).

%% mapl(F,List)
%% an alternative map which always maps from left to right
%% and makes it possible to interrupt the mapping with throw on
%% the first occurence from left as expected.
%% can be removed when the jam machine (and all other machines)
%% uses the standardized (Erlang 5.0) evaluation order (from left to right)
mapl(F, [H|T]) ->
	V = F(H),
	[V | mapl(F,T)];
mapl(_, []) ->
	[].

%% normalise(AbsTerm)
%% abstract(Term)
%%  Convert between the abstract form of a term and a term.

normalise({char,_,C}) -> C;
normalise({integer,_,I}) -> I;
normalise({float,_,F}) -> F;
normalise({atom,_,A}) -> A;
normalise({string,_,S}) -> S;
normalise({nil,_}) -> [];
normalise({bin,_,Fs}) ->
    {value, B, _} =
	eval_bits:expr_grp(Fs, [],
			   fun(E, _) ->
				   {value, normalise(E), []}
			   end, [], true),
    B;
normalise({cons,_,Head,Tail}) ->
    [normalise(Head)|normalise(Tail)];
normalise({tuple,_,Args}) ->
    list_to_tuple(normalise_list(Args));
%% Atom dot-notation, as in 'foo.bar.baz'
normalise({record_field,_,_,_}=A) ->
    case package_segments(A) of
	error -> erlang:fault({badarg, A});
	As -> list_to_atom(packages:concat(As))
    end;
%% Special case for unary +/-.
normalise({op,_,'+',{char,_,I}}) -> I;
normalise({op,_,'+',{integer,_,I}}) -> I;
normalise({op,_,'+',{float,_,F}}) -> F;
normalise({op,_,'-',{char,_,I}}) -> -I;		%Weird, but compatible!
normalise({op,_,'-',{integer,_,I}}) -> -I;
normalise({op,_,'-',{float,_,F}}) -> -F;
normalise(X) -> erlang:fault({badarg, X}).

normalise_list([H|T]) ->
    [normalise(H)|normalise_list(T)];
normalise_list([]) ->
    [].

abstract(T) when integer(T) -> {integer,0,T};
abstract(T) when float(T) -> {float,0,T};
abstract(T) when atom(T) -> {atom,0,T};
abstract([]) -> {nil,0};
abstract(B) when binary(B) ->
    {bin, 0, lists:map(fun(Byte) ->
			       {bin_element, 0,
				{integer, 0, Byte}, default, default}
		       end,
		       binary_to_list(B))};
abstract([C|T]) when integer(C), 0 =< C, C < 256 ->
    abstract_string(T, [C]);
abstract([H|T]) ->
    {cons,0,abstract(H),abstract(T)};
abstract(Tuple) when tuple(Tuple) ->
    {tuple,0,abstract_list(tuple_to_list(Tuple))}.

abstract_string([C|T], String) when integer(C), 0 =< C, C < 256 ->
    abstract_string(T, [C|String]);
abstract_string([], String) ->
    {string, 0, lists:reverse(String)};
abstract_string(T, String) ->
    not_string(String, abstract(T)).

not_string([C|T], Result) ->
    not_string(T, {cons, 0, {integer, 0, C}, Result});
not_string([], Result) ->
    Result.

abstract_list([H|T]) ->
    [abstract(H)|abstract_list(T)];
abstract_list([]) ->
    [].

%%% abstract/2 keeps the line number
abstract(T, Line) when integer(T) -> {integer,Line,T};
abstract(T, Line) when float(T) -> {float,Line,T};
abstract(T, Line) when atom(T) -> {atom,Line,T};
abstract([], Line) -> {nil,Line};
abstract(B, Line) when binary(B) ->
    {bin, Line, lists:map(fun(Byte) ->
			       {bin_element, Line,
				{integer, Line, Byte}, default, default}
		       end,
		       binary_to_list(B))};
abstract([C|T], Line) when integer(C), 0 =< C, C < 256 ->
    abstract_string(T, [C], Line);
abstract([H|T], Line) ->
    {cons,Line,abstract(H, Line),abstract(T, Line)};
abstract(Tuple, Line) when tuple(Tuple) ->
    {tuple,Line,abstract_list(tuple_to_list(Tuple), Line)}.

abstract_string([C|T], String, Line) when integer(C), 0 =< C, C < 256 ->
    abstract_string(T, [C|String], Line);
abstract_string([], String, Line) ->
    {string, Line, lists:reverse(String)};
abstract_string(T, String, Line) ->
    not_string(String, abstract(T, Line), Line).

not_string([C|T], Result, Line) ->
    not_string(T, {cons, Line, {integer, Line, C}, Result}, Line);
not_string([], Result, Line) ->
    Result.

abstract_list([H|T], Line) ->
    [abstract(H, Line)|abstract_list(T, Line)];
abstract_list([], Line) ->
    [].

%% tokens(AbsTerm) -> [Token]
%% tokens(AbsTerm, More) -> [Token]
%%  Generate a list of tokens representing the abstract term.

tokens(Abs) ->
    tokens(Abs, []).

tokens({char,L,C}, More) -> [{char,L,C}|More];
tokens({integer,L,N}, More) -> [{integer,L,N}|More];
tokens({float,L,F}, More) -> [{float,L,F}|More];
tokens({atom,L,A}, More) -> [{atom,L,A}|More];
tokens({var,L,V}, More) -> [{var,L,V}|More];
tokens({string,L,S}, More) -> [{string,L,S}|More];
tokens({nil,L}, More) -> [{'[',L},{']',L}|More];
tokens({cons,L,Head,Tail}, More) ->
    [{'[',L}|tokens(Head, tokens_tail(Tail, More))];
tokens({tuple,L,[]}, More) ->
    [{'{',L},{'}',L}|More];
tokens({tuple,L,[E|Es]}, More) ->
    [{'{',L}|tokens(E, tokens_tuple(Es, line(E), More))].

tokens_tail({cons,L,Head,Tail}, More) ->
    [{',',L}|tokens(Head, tokens_tail(Tail, More))];
tokens_tail({nil,L}, More) ->
    [{']',L}|More];
tokens_tail(Other, More) ->
    L = line(Other),
    [{'|',L}|tokens(Other, [{']',L}|More])].

tokens_tuple([E|Es], Line, More) ->
    [{',',Line}|tokens(E, tokens_tuple(Es, line(E), More))];
tokens_tuple([], Line, More) ->
    [{'}',Line}|More].

%% Give the relative precedences of operators.

inop_prec('=') -> {150,100,100};
inop_prec('!') -> {150,100,100};
inop_prec('orelse') -> {160,150,150};
inop_prec('andalso') -> {200,160,160};
inop_prec('==') -> {300,200,300};
inop_prec('/=') -> {300,200,300};
inop_prec('=<') -> {300,200,300};
inop_prec('<') -> {300,200,300};
inop_prec('>=') -> {300,200,300};
inop_prec('>') -> {300,200,300};
inop_prec('=:=') -> {300,200,300};
inop_prec('=/=') -> {300,200,300};
inop_prec('++') -> {300,300,400};
inop_prec('--') -> {300,300,400};
inop_prec('+') -> {400,400,500};
inop_prec('-') -> {400,400,500};
inop_prec('bor') -> {400,400,500};
inop_prec('bxor') -> {400,400,500};
inop_prec('bsl') -> {400,400,500};
inop_prec('bsr') -> {400,400,500};
inop_prec('or') -> {400,400,500};
inop_prec('xor') -> {400,400,500};
inop_prec('*') -> {500,500,600};
inop_prec('/') -> {500,500,600};
inop_prec('div') -> {500,500,600};
inop_prec('rem') -> {500,500,600};
inop_prec('band') -> {500,500,600};
inop_prec('and') -> {500,500,600};
inop_prec('#') -> {800,700,800};
inop_prec(':') -> {900,800,900};
inop_prec('.') -> {900,900,1000}.

preop_prec('catch') -> {0,100};
preop_prec('+') -> {600,700};
preop_prec('-') -> {600,700};
preop_prec('bnot') -> {600,700};
preop_prec('not') -> {600,700};
preop_prec('#') -> {700,800}.

func_prec() -> {800,700}.

max_prec() -> 1000.

%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% The parser generator will insert appropriate declarations before this line.%

parse(Tokens) ->
    case catch yeccpars1(Tokens, false, 0, [], []) of
	error ->
	    Errorline =
		if Tokens == [] -> 0; true -> element(2, hd(Tokens)) end,
	    {error,
	     {Errorline, ?THIS_MODULE, "syntax error at or after this line."}};
	Other ->
	    Other
    end.

parse_and_scan({Mod, Fun, Args}) ->
    case apply(Mod, Fun, Args) of
	{eof, _} ->
	    {ok, eof};
	{error, Descriptor, _} ->
	    {error, Descriptor};
	{ok, Tokens, _} ->
	    yeccpars1(Tokens, {Mod, Fun, Args}, 0, [], [])
    end.

format_error(Message) ->
    case io_lib:deep_char_list(Message) of
	true ->
	    Message;
	_ ->
	    io_lib:write(Message)
    end.

% To be used in grammar files to throw an error message to the parser toplevel.
% Doesn't have to be exported!
return_error(Line, Message) ->
    throw({error, {Line, ?THIS_MODULE, Message}}).


% Don't change yeccpars1/6 too much, it is called recursively by yeccpars2/8!
yeccpars1([Token | Tokens], Tokenizer, State, States, Vstack) ->
    yeccpars2(State, element(1, Token), States, Vstack, Token, Tokens,
	      Tokenizer);
yeccpars1([], {M, F, A}, State, States, Vstack) ->
    case catch apply(M, F, A) of
        {eof, Endline} ->
            {error, {Endline, ?THIS_MODULE, "end_of_file"}};
        {error, Descriptor, _Endline} ->
            {error, Descriptor};
        {'EXIT', Reason} ->
            {error, {0, ?THIS_MODULE, Reason}};
        {ok, Tokens, _Endline} ->
	    case catch yeccpars1(Tokens, {M, F, A}, State, States, Vstack) of
		error ->
		    Errorline = element(2, hd(Tokens)),
		    {error, {Errorline, ?THIS_MODULE,
			     "syntax error at or after this line."}};
		Other ->
		    Other
	    end
    end;
yeccpars1([], false, State, States, Vstack) ->
    yeccpars2(State, '$end', States, Vstack, {'$end', 999999}, [], false).

% For internal use only.
yeccerror(Token) ->
    {error,
     {element(2, Token), ?THIS_MODULE,
      ["syntax error before: ", yecctoken2string(Token)]}}.

yecctoken2string({atom, _, A}) -> io_lib:write(A);
yecctoken2string({integer,_,N}) -> io_lib:write(N);
yecctoken2string({float,_,F}) -> io_lib:write(F);
yecctoken2string({char,_,C}) -> io_lib:write_char(C);
yecctoken2string({var,_,V}) -> io_lib:format('~s', [V]);
yecctoken2string({string,_,S}) -> io_lib:write_string(S);
yecctoken2string({reserved_symbol, _, A}) -> io_lib:format('~w', [A]);
yecctoken2string({_Cat, _, Val}) -> io_lib:format('~w', [Val]);

yecctoken2string({'dot', _}) -> io_lib:format('~w', ['.']);
yecctoken2string({'$end', _}) ->
    [];
yecctoken2string({Other, _}) when atom(Other) ->
    io_lib:format('~w', [Other]);
yecctoken2string(Other) ->
    io_lib:write(Other).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


yeccpars2(0, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 2, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(1, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 256, [1 | __Ss], [__T | __Stack]);
yeccpars2(1, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(2, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [2 | __Ss], [__T | __Stack]);
yeccpars2(2, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(3, 'dot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 253, [3 | __Ss], [__T | __Stack]);
yeccpars2(3, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(4, '$end', _, __Stack, _, _, _) ->
 {ok, hd(__Stack)};
yeccpars2(4, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(5, 'dot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 252, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(6, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 246, [6 | __Ss], [__T | __Stack]);
yeccpars2(6, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(function_clauses, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(7, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  build_function(__1),
 yeccpars2(yeccgoto(function, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(8, 'dot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 245, [8 | __Ss], [__T | __Stack]);
yeccpars2(8, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(9, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [9 | __Ss], [__T | __Stack]);
yeccpars2(9, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(rule_clauses, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(10, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  build_rule(__1),
 yeccpars2(yeccgoto(rule, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(11, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [11 | __Ss], [__T | __Stack]);
yeccpars2(11, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(12, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [12 | __Ss], [__T | __Stack]);
yeccpars2(12, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(13, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(rule_clauses, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(14, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 242, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(15, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(1,__1),
 yeccpars2(yeccgoto(clause_args, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(16, 'when', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [16 | __Ss], [__T | __Stack]);
yeccpars2(16, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(17, __Cat, [16 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(17, ':-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 239, [17 | __Ss], [__T | __Stack]);
yeccpars2(17, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(18, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(19, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 235, [19 | __Ss], [__T | __Stack]);
yeccpars2(19, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(20, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(21, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(prefix_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(22, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(prefix_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(23, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 232, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(24, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '>>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 209, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(25, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 199, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(26, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(atomic, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(27, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(28, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(29, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(30, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(prefix_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(31, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [31 | __Ss], [__T | __Stack]);
yeccpars2(31, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(32, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(33, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(34, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(atomic, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(35, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 190, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(exprs, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(36, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(37, '!', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 186, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, '=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 187, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_100, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(38, 'orelse', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 184, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_150, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(39, 'andalso', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 182, [39 | __Ss], [__T | __Stack]);
yeccpars2(39, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_160, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(40, '=/=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 174, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '=:=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 175, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 178, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '>=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 179, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 173, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '=<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 176, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '/=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 172, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '==', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 177, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_200, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(41, '--', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 161, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, '++', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 159, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, 'xor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 169, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, 'or', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 168, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, 'bsr', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 165, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, 'bsl', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 164, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, 'bxor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 166, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, 'bor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 163, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 160, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 158, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_300, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(42, 'and', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 152, [42 | __Ss], [__T | __Stack]);
yeccpars2(42, 'band', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 153, [42 | __Ss], [__T | __Stack]);
yeccpars2(42, 'rem', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 156, [42 | __Ss], [__T | __Stack]);
yeccpars2(42, 'div', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 154, [42 | __Ss], [__T | __Stack]);
yeccpars2(42, '*', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 150, [42 | __Ss], [__T | __Stack]);
yeccpars2(42, '/', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 151, [42 | __Ss], [__T | __Stack]);
yeccpars2(42, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_400, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(43, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_500, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(44, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_600, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(45, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [45 | __Ss], [__T | __Stack]);
yeccpars2(45, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_700, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(46, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 145, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_800, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(47, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 127, [47 | __Ss], [__T | __Stack]);
yeccpars2(47, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_900, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(48, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 125, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(guard, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(49, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(atomic, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(50, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [50 | __Ss], [__T | __Stack]);
yeccpars2(50, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [50 | __Ss], [__T | __Stack]);
yeccpars2(50, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(51, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(52, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_700, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(53, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(clause_guard, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(54, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [54 | __Ss], [__T | __Stack]);
yeccpars2(54, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(55, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(56, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(atomic, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(57, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(58, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(59, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(prefix_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(60, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(61, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 93, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(62, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(63, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'after', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(64, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(65, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_700, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(66, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(strings, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(67, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(atomic, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(68, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(69, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(70, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(71, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [71 | __Ss], [__T | __Stack]);
yeccpars2(71, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(72, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {tuple,line(__1),[]},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(tuple, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(73, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {tuple,line(__1),__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tuple, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(74, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {string,line(__1),element(3,__1) ++ element(3,__2)},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(strings, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(75, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(76, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [76 | __Ss], [__T | __Stack]);
yeccpars2(76, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(cr_clauses, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(77, 'after', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(78, 'when', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [78 | __Ss], [__T | __Stack]);
yeccpars2(78, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(79, __Cat, [78 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(79, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(80, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(81, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {clause,line(__1),[__1],__2,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(cr_clause, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(82, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(clause_body, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(83, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(84, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'receive',line(__1),__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(receive_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(85, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(86, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(87, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'receive',line(__1),__2,__4,__5},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(receive_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(88, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(89, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(cr_clauses, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(90, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(91, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 92, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(92, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'receive',line(__1),[],__3,__4},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(receive_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(93, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(94, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 95, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(95, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'query',line(__1),__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(query_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(96, '||', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(97, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(98, '<-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(lc_expr, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(99, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(lc_exprs, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(100, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 101, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(101, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {lc,line(__1),__2,__4},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(list_comprehension, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(102, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(103, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(lc_exprs, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(104, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(105, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {generate,line(__2),__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(lc_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(106, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(expr_600, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(107, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(108, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 111, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(if_clauses, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(109, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 110, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(110, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'if',line(__1),__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(if_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(111, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [111 | __Ss], [__T | __Stack]);
yeccpars2(111, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(112, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(if_clauses, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(113, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {clause,line(hd(hd(__1))),[],__1,__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(if_clause, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(114, 'when', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(123, __Cat, [114 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(115, '/', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 121, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(116, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 119, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(fun_clauses, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(117, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 118, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(118, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  build_fun(line(__1),__2),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(fun_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(119, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [119 | __Ss], [__T | __Stack]);
yeccpars2(119, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(120, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(fun_clauses, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(121, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [121 | __Ss], [__T | __Stack]);
yeccpars2(121, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(122, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'fun',line(__1),{function,element(3,__2),element(3,__4)}},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(fun_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(123, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [123 | __Ss], [__T | __Stack]);
yeccpars2(123, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(124, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   {Args,Pos} = __1, {clause,Pos,'fun',Args,__2,__3}
  end,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(fun_clause, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(125, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(126, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(guard, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(127, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 128, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(128, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 131, [128 | __Ss], [__T | __Stack]);
yeccpars2(128, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 129, [128 | __Ss], [__T | __Stack]);
yeccpars2(128, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(129, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(130, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {record,line(__2),__1,element(3,__3),__4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(record_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(131, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 132, [131 | __Ss], [__T | __Stack]);
yeccpars2(131, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 135, [131 | __Ss], [__T | __Stack]);
yeccpars2(131, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 136, [131 | __Ss], [__T | __Stack]);
yeccpars2(131, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(132, '=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(133, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [133 | __Ss], [__T | __Stack]);
yeccpars2(133, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(record_fields, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(134, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 139, [134 | __Ss], [__T | __Stack]);
yeccpars2(134, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(135, '=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [135 | __Ss], [__T | __Stack]);
yeccpars2(135, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(136, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(record_tuple, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(137, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(138, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {record_field,line(__1),__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(record_field, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(139, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(record_tuple, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(140, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 132, [140 | __Ss], [__T | __Stack]);
yeccpars2(140, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 135, [140 | __Ss], [__T | __Stack]);
yeccpars2(140, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(141, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(record_fields, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(142, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(143, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {record_field,line(__1),__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(record_field, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(144, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {record_field,line(__2),__1,element(3,__3),__5},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(record_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(145, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(146, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(147, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {remote,line(__2),__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_800, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(148, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {record_field,line(__2),__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_900, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(149, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {call,line(__1),__1,element(1,__2)},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(function_call, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(150, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(mult_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(151, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(mult_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(152, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(mult_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(153, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(mult_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(154, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(mult_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(155, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(156, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(mult_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(157, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_500, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(158, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(add_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(159, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(list_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(160, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(add_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(161, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(list_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(162, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(163, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(add_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(164, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(add_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(165, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(add_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(166, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(add_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(167, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(168, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(add_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(169, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(add_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(170, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_300, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(171, 'and', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 152, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, 'band', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 153, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, 'rem', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 156, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, 'div', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 154, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, '*', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 150, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, '/', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 151, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_400, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(172, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(comp_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(173, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(comp_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(174, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(comp_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(175, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(comp_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(176, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(comp_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(177, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(comp_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(178, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(comp_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(179, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(comp_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(180, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(181, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_200, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(182, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(183, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_160, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(184, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(185, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_150, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(186, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(187, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(188, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {match,line(__2),__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_100, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(189, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_100, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(190, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(191, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(exprs, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(192, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'catch',line(__1),__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(193, 'of', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 194, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(194, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(195, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 196, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(196, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'case',line(__1),__2,__4},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(case_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(197, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 198, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(198, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {block,line(__1),__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_max, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(199, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {nil,line(__1)},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(list, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(200, '||', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 201, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 204, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 202, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(201, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(202, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {nil,line(__1)},
 yeccpars2(yeccgoto(tail, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(203, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {cons,line(__1),__2,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(list, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(204, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(205, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 206, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(206, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tail, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(207, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 201, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 204, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 202, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(208, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {cons,line(__2),__2,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tail, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(209, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {bin,line(__1),[]},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(binary, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(210, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 230, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(bin_elements, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(211, '>>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 229, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(212, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 216, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  default,
 yeccpars2(217, __Cat, [212 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(213, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(bit_expr, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(214, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [214 | __Ss], [__T | __Stack]);
yeccpars2(214, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(215, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(bit_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(216, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(217, '/', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 218, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  default,
 yeccpars2(219, __Cat, [217 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(218, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 220, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(219, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {bin_element,line(__1),__1,__2,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(bin_element, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(220, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 225, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(3,__1),
 yeccpars2(yeccgoto(bit_type, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(221, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 223, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(bit_type_list, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(222, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(opt_bit_type_list, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(223, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 220, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(224, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(bit_type_list, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(225, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 226, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(226, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {element(3,__1),element(3,__3)},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(bit_type, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(227, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(opt_bit_size_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(228, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(bit_size_expr, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(229, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {bin,line(__1),__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(binary, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(230, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(231, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(bin_elements, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(232, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {record_field,line(__1),{atom,line(__1),''},__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(expr_900, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(233, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 234, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(234, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_max, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(235, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 131, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 236, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(236, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 238, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(237, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {record,line(__1),element(3,__2),__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(record_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(238, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {record_index,line(__1),element(3,__2),__4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(record_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(239, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(240, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {clause,line(__1),element(3,__1),__2,__3,__4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(rule_clause, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(241, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(rule_body, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(242, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {[],line(__1)},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(argument_list, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(243, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 244, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(244, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__2,line(__1)},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(argument_list, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(245, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(form, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(246, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 247, [246 | __Ss], [__T | __Stack]);
yeccpars2(246, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(247, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [247 | __Ss], [__T | __Stack]);
yeccpars2(247, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(248, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(function_clauses, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(249, 'when', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [249 | __Ss], [__T | __Stack]);
yeccpars2(249, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(250, __Cat, [249 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(250, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [250 | __Ss], [__T | __Stack]);
yeccpars2(250, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(251, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {clause,line(__1),element(3,__1),__2,__3,__4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(function_clause, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(252, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(form, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(253, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(form, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(254, 'when', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [254 | __Ss], [__T | __Stack]);
yeccpars2(254, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(255, __Cat, [254 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(255, ':-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 239, [255 | __Ss], [__T | __Stack]);
yeccpars2(255, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [255 | __Ss], [__T | __Stack]);
yeccpars2(255, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(256, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 257, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(257, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(258, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 260, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(259, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(attr_val, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(260, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  build_attribute(__2,__4),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(attribute, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(__Other, _, _, _, _, _, _) ->
 exit({parser, __Other, missing_state_in_action_table}).

yeccgoto(add_op, 41) ->
 162;
yeccgoto(argument_list, 2) ->
 15;
yeccgoto(argument_list, 12) ->
 15;
yeccgoto(argument_list, 45) ->
 149;
yeccgoto(argument_list, 50) ->
 114;
yeccgoto(argument_list, 119) ->
 114;
yeccgoto(argument_list, 247) ->
 15;
yeccgoto(atomic, 14) ->
 27;
yeccgoto(atomic, 18) ->
 27;
yeccgoto(atomic, 20) ->
 27;
yeccgoto(atomic, 24) ->
 27;
yeccgoto(atomic, 25) ->
 27;
yeccgoto(atomic, 28) ->
 27;
yeccgoto(atomic, 31) ->
 27;
yeccgoto(atomic, 33) ->
 27;
yeccgoto(atomic, 54) ->
 27;
yeccgoto(atomic, 60) ->
 27;
yeccgoto(atomic, 63) ->
 27;
yeccgoto(atomic, 70) ->
 27;
yeccgoto(atomic, 75) ->
 27;
yeccgoto(atomic, 80) ->
 27;
yeccgoto(atomic, 83) ->
 27;
yeccgoto(atomic, 88) ->
 27;
yeccgoto(atomic, 93) ->
 27;
yeccgoto(atomic, 97) ->
 27;
yeccgoto(atomic, 102) ->
 27;
yeccgoto(atomic, 104) ->
 27;
yeccgoto(atomic, 111) ->
 27;
yeccgoto(atomic, 125) ->
 27;
yeccgoto(atomic, 137) ->
 27;
yeccgoto(atomic, 142) ->
 27;
yeccgoto(atomic, 146) ->
 27;
yeccgoto(atomic, 155) ->
 27;
yeccgoto(atomic, 162) ->
 27;
yeccgoto(atomic, 167) ->
 27;
yeccgoto(atomic, 180) ->
 27;
yeccgoto(atomic, 182) ->
 27;
yeccgoto(atomic, 184) ->
 27;
yeccgoto(atomic, 186) ->
 27;
yeccgoto(atomic, 187) ->
 27;
yeccgoto(atomic, 190) ->
 27;
yeccgoto(atomic, 194) ->
 27;
yeccgoto(atomic, 201) ->
 27;
yeccgoto(atomic, 204) ->
 27;
yeccgoto(atomic, 214) ->
 27;
yeccgoto(atomic, 216) ->
 27;
yeccgoto(atomic, 230) ->
 27;
yeccgoto(atomic, 239) ->
 27;
yeccgoto(atomic, 257) ->
 27;
yeccgoto(attr_val, 257) ->
 258;
yeccgoto(attribute, 0) ->
 3;
yeccgoto(bin_element, 24) ->
 210;
yeccgoto(bin_element, 230) ->
 210;
yeccgoto(bin_elements, 24) ->
 211;
yeccgoto(bin_elements, 230) ->
 231;
yeccgoto(binary, 14) ->
 29;
yeccgoto(binary, 18) ->
 29;
yeccgoto(binary, 20) ->
 29;
yeccgoto(binary, 24) ->
 29;
yeccgoto(binary, 25) ->
 29;
yeccgoto(binary, 28) ->
 29;
yeccgoto(binary, 31) ->
 29;
yeccgoto(binary, 33) ->
 29;
yeccgoto(binary, 54) ->
 29;
yeccgoto(binary, 60) ->
 29;
yeccgoto(binary, 63) ->
 29;
yeccgoto(binary, 70) ->
 29;
yeccgoto(binary, 75) ->
 29;
yeccgoto(binary, 80) ->
 29;
yeccgoto(binary, 83) ->
 29;
yeccgoto(binary, 88) ->
 29;
yeccgoto(binary, 93) ->
 29;
yeccgoto(binary, 97) ->
 29;
yeccgoto(binary, 102) ->
 29;
yeccgoto(binary, 104) ->
 29;
yeccgoto(binary, 111) ->
 29;
yeccgoto(binary, 125) ->
 29;
yeccgoto(binary, 137) ->
 29;
yeccgoto(binary, 142) ->
 29;
yeccgoto(binary, 146) ->
 29;
yeccgoto(binary, 155) ->
 29;
yeccgoto(binary, 162) ->
 29;
yeccgoto(binary, 167) ->
 29;
yeccgoto(binary, 180) ->
 29;
yeccgoto(binary, 182) ->
 29;
yeccgoto(binary, 184) ->
 29;
yeccgoto(binary, 186) ->
 29;
yeccgoto(binary, 187) ->
 29;
yeccgoto(binary, 190) ->
 29;
yeccgoto(binary, 194) ->
 29;
yeccgoto(binary, 201) ->
 29;
yeccgoto(binary, 204) ->
 29;
yeccgoto(binary, 214) ->
 29;
yeccgoto(binary, 216) ->
 29;
yeccgoto(binary, 230) ->
 29;
yeccgoto(binary, 239) ->
 29;
yeccgoto(binary, 257) ->
 29;
yeccgoto(bit_expr, 24) ->
 212;
yeccgoto(bit_expr, 230) ->
 212;
yeccgoto(bit_size_expr, 216) ->
 227;
yeccgoto(bit_type, 218) ->
 221;
yeccgoto(bit_type, 223) ->
 221;
yeccgoto(bit_type_list, 218) ->
 222;
yeccgoto(bit_type_list, 223) ->
 224;
yeccgoto(case_expr, 14) ->
 32;
yeccgoto(case_expr, 18) ->
 32;
yeccgoto(case_expr, 20) ->
 32;
yeccgoto(case_expr, 24) ->
 32;
yeccgoto(case_expr, 25) ->
 32;
yeccgoto(case_expr, 28) ->
 32;
yeccgoto(case_expr, 31) ->
 32;
yeccgoto(case_expr, 33) ->
 32;
yeccgoto(case_expr, 54) ->
 32;
yeccgoto(case_expr, 60) ->
 32;
yeccgoto(case_expr, 63) ->
 32;
yeccgoto(case_expr, 70) ->
 32;
yeccgoto(case_expr, 75) ->
 32;
yeccgoto(case_expr, 80) ->
 32;
yeccgoto(case_expr, 83) ->
 32;
yeccgoto(case_expr, 88) ->
 32;
yeccgoto(case_expr, 93) ->
 32;
yeccgoto(case_expr, 97) ->
 32;
yeccgoto(case_expr, 102) ->
 32;
yeccgoto(case_expr, 104) ->
 32;
yeccgoto(case_expr, 111) ->
 32;
yeccgoto(case_expr, 125) ->
 32;
yeccgoto(case_expr, 137) ->
 32;
yeccgoto(case_expr, 142) ->
 32;
yeccgoto(case_expr, 146) ->
 32;
yeccgoto(case_expr, 155) ->
 32;
yeccgoto(case_expr, 162) ->
 32;
yeccgoto(case_expr, 167) ->
 32;
yeccgoto(case_expr, 180) ->
 32;
yeccgoto(case_expr, 182) ->
 32;
yeccgoto(case_expr, 184) ->
 32;
yeccgoto(case_expr, 186) ->
 32;
yeccgoto(case_expr, 187) ->
 32;
yeccgoto(case_expr, 190) ->
 32;
yeccgoto(case_expr, 194) ->
 32;
yeccgoto(case_expr, 201) ->
 32;
yeccgoto(case_expr, 204) ->
 32;
yeccgoto(case_expr, 214) ->
 32;
yeccgoto(case_expr, 216) ->
 32;
yeccgoto(case_expr, 230) ->
 32;
yeccgoto(case_expr, 239) ->
 32;
yeccgoto(case_expr, 257) ->
 32;
yeccgoto(clause_args, 2) ->
 254;
yeccgoto(clause_args, 12) ->
 16;
yeccgoto(clause_args, 247) ->
 249;
yeccgoto(clause_body, 79) ->
 81;
yeccgoto(clause_body, 85) ->
 86;
yeccgoto(clause_body, 90) ->
 91;
yeccgoto(clause_body, 107) ->
 113;
yeccgoto(clause_body, 123) ->
 124;
yeccgoto(clause_body, 250) ->
 251;
yeccgoto(clause_body, 255) ->
 251;
yeccgoto(clause_guard, 16) ->
 17;
yeccgoto(clause_guard, 78) ->
 79;
yeccgoto(clause_guard, 114) ->
 123;
yeccgoto(clause_guard, 249) ->
 250;
yeccgoto(clause_guard, 254) ->
 255;
yeccgoto(comp_op, 40) ->
 180;
yeccgoto(cr_clause, 63) ->
 76;
yeccgoto(cr_clause, 88) ->
 76;
yeccgoto(cr_clause, 194) ->
 76;
yeccgoto(cr_clauses, 63) ->
 77;
yeccgoto(cr_clauses, 88) ->
 89;
yeccgoto(cr_clauses, 194) ->
 195;
yeccgoto(expr, 14) ->
 35;
yeccgoto(expr, 18) ->
 35;
yeccgoto(expr, 20) ->
 233;
yeccgoto(expr, 25) ->
 200;
yeccgoto(expr, 28) ->
 35;
yeccgoto(expr, 31) ->
 193;
yeccgoto(expr, 33) ->
 192;
yeccgoto(expr, 54) ->
 35;
yeccgoto(expr, 63) ->
 78;
yeccgoto(expr, 70) ->
 35;
yeccgoto(expr, 75) ->
 90;
yeccgoto(expr, 80) ->
 35;
yeccgoto(expr, 83) ->
 85;
yeccgoto(expr, 88) ->
 78;
yeccgoto(expr, 93) ->
 96;
yeccgoto(expr, 97) ->
 98;
yeccgoto(expr, 102) ->
 98;
yeccgoto(expr, 104) ->
 105;
yeccgoto(expr, 111) ->
 35;
yeccgoto(expr, 125) ->
 35;
yeccgoto(expr, 137) ->
 138;
yeccgoto(expr, 142) ->
 143;
yeccgoto(expr, 190) ->
 35;
yeccgoto(expr, 194) ->
 78;
yeccgoto(expr, 201) ->
 207;
yeccgoto(expr, 204) ->
 205;
yeccgoto(expr, 239) ->
 98;
yeccgoto(expr, 257) ->
 35;
yeccgoto(expr_100, 14) ->
 36;
yeccgoto(expr_100, 18) ->
 36;
yeccgoto(expr_100, 20) ->
 36;
yeccgoto(expr_100, 25) ->
 36;
yeccgoto(expr_100, 28) ->
 36;
yeccgoto(expr_100, 31) ->
 36;
yeccgoto(expr_100, 33) ->
 36;
yeccgoto(expr_100, 54) ->
 36;
yeccgoto(expr_100, 63) ->
 36;
yeccgoto(expr_100, 70) ->
 36;
yeccgoto(expr_100, 75) ->
 36;
yeccgoto(expr_100, 80) ->
 36;
yeccgoto(expr_100, 83) ->
 36;
yeccgoto(expr_100, 88) ->
 36;
yeccgoto(expr_100, 93) ->
 36;
yeccgoto(expr_100, 97) ->
 36;
yeccgoto(expr_100, 102) ->
 36;
yeccgoto(expr_100, 104) ->
 36;
yeccgoto(expr_100, 111) ->
 36;
yeccgoto(expr_100, 125) ->
 36;
yeccgoto(expr_100, 137) ->
 36;
yeccgoto(expr_100, 142) ->
 36;
yeccgoto(expr_100, 186) ->
 189;
yeccgoto(expr_100, 187) ->
 188;
yeccgoto(expr_100, 190) ->
 36;
yeccgoto(expr_100, 194) ->
 36;
yeccgoto(expr_100, 201) ->
 36;
yeccgoto(expr_100, 204) ->
 36;
yeccgoto(expr_100, 239) ->
 36;
yeccgoto(expr_100, 257) ->
 36;
yeccgoto(expr_150, 14) ->
 37;
yeccgoto(expr_150, 18) ->
 37;
yeccgoto(expr_150, 20) ->
 37;
yeccgoto(expr_150, 25) ->
 37;
yeccgoto(expr_150, 28) ->
 37;
yeccgoto(expr_150, 31) ->
 37;
yeccgoto(expr_150, 33) ->
 37;
yeccgoto(expr_150, 54) ->
 37;
yeccgoto(expr_150, 63) ->
 37;
yeccgoto(expr_150, 70) ->
 37;
yeccgoto(expr_150, 75) ->
 37;
yeccgoto(expr_150, 80) ->
 37;
yeccgoto(expr_150, 83) ->
 37;
yeccgoto(expr_150, 88) ->
 37;
yeccgoto(expr_150, 93) ->
 37;
yeccgoto(expr_150, 97) ->
 37;
yeccgoto(expr_150, 102) ->
 37;
yeccgoto(expr_150, 104) ->
 37;
yeccgoto(expr_150, 111) ->
 37;
yeccgoto(expr_150, 125) ->
 37;
yeccgoto(expr_150, 137) ->
 37;
yeccgoto(expr_150, 142) ->
 37;
yeccgoto(expr_150, 184) ->
 185;
yeccgoto(expr_150, 186) ->
 37;
yeccgoto(expr_150, 187) ->
 37;
yeccgoto(expr_150, 190) ->
 37;
yeccgoto(expr_150, 194) ->
 37;
yeccgoto(expr_150, 201) ->
 37;
yeccgoto(expr_150, 204) ->
 37;
yeccgoto(expr_150, 239) ->
 37;
yeccgoto(expr_150, 257) ->
 37;
yeccgoto(expr_160, 14) ->
 38;
yeccgoto(expr_160, 18) ->
 38;
yeccgoto(expr_160, 20) ->
 38;
yeccgoto(expr_160, 25) ->
 38;
yeccgoto(expr_160, 28) ->
 38;
yeccgoto(expr_160, 31) ->
 38;
yeccgoto(expr_160, 33) ->
 38;
yeccgoto(expr_160, 54) ->
 38;
yeccgoto(expr_160, 63) ->
 38;
yeccgoto(expr_160, 70) ->
 38;
yeccgoto(expr_160, 75) ->
 38;
yeccgoto(expr_160, 80) ->
 38;
yeccgoto(expr_160, 83) ->
 38;
yeccgoto(expr_160, 88) ->
 38;
yeccgoto(expr_160, 93) ->
 38;
yeccgoto(expr_160, 97) ->
 38;
yeccgoto(expr_160, 102) ->
 38;
yeccgoto(expr_160, 104) ->
 38;
yeccgoto(expr_160, 111) ->
 38;
yeccgoto(expr_160, 125) ->
 38;
yeccgoto(expr_160, 137) ->
 38;
yeccgoto(expr_160, 142) ->
 38;
yeccgoto(expr_160, 182) ->
 183;
yeccgoto(expr_160, 184) ->
 38;
yeccgoto(expr_160, 186) ->
 38;
yeccgoto(expr_160, 187) ->
 38;
yeccgoto(expr_160, 190) ->
 38;
yeccgoto(expr_160, 194) ->
 38;
yeccgoto(expr_160, 201) ->
 38;
yeccgoto(expr_160, 204) ->
 38;
yeccgoto(expr_160, 239) ->
 38;
yeccgoto(expr_160, 257) ->
 38;
yeccgoto(expr_200, 14) ->
 39;
yeccgoto(expr_200, 18) ->
 39;
yeccgoto(expr_200, 20) ->
 39;
yeccgoto(expr_200, 25) ->
 39;
yeccgoto(expr_200, 28) ->
 39;
yeccgoto(expr_200, 31) ->
 39;
yeccgoto(expr_200, 33) ->
 39;
yeccgoto(expr_200, 54) ->
 39;
yeccgoto(expr_200, 63) ->
 39;
yeccgoto(expr_200, 70) ->
 39;
yeccgoto(expr_200, 75) ->
 39;
yeccgoto(expr_200, 80) ->
 39;
yeccgoto(expr_200, 83) ->
 39;
yeccgoto(expr_200, 88) ->
 39;
yeccgoto(expr_200, 93) ->
 39;
yeccgoto(expr_200, 97) ->
 39;
yeccgoto(expr_200, 102) ->
 39;
yeccgoto(expr_200, 104) ->
 39;
yeccgoto(expr_200, 111) ->
 39;
yeccgoto(expr_200, 125) ->
 39;
yeccgoto(expr_200, 137) ->
 39;
yeccgoto(expr_200, 142) ->
 39;
yeccgoto(expr_200, 182) ->
 39;
yeccgoto(expr_200, 184) ->
 39;
yeccgoto(expr_200, 186) ->
 39;
yeccgoto(expr_200, 187) ->
 39;
yeccgoto(expr_200, 190) ->
 39;
yeccgoto(expr_200, 194) ->
 39;
yeccgoto(expr_200, 201) ->
 39;
yeccgoto(expr_200, 204) ->
 39;
yeccgoto(expr_200, 239) ->
 39;
yeccgoto(expr_200, 257) ->
 39;
yeccgoto(expr_300, 14) ->
 40;
yeccgoto(expr_300, 18) ->
 40;
yeccgoto(expr_300, 20) ->
 40;
yeccgoto(expr_300, 25) ->
 40;
yeccgoto(expr_300, 28) ->
 40;
yeccgoto(expr_300, 31) ->
 40;
yeccgoto(expr_300, 33) ->
 40;
yeccgoto(expr_300, 54) ->
 40;
yeccgoto(expr_300, 63) ->
 40;
yeccgoto(expr_300, 70) ->
 40;
yeccgoto(expr_300, 75) ->
 40;
yeccgoto(expr_300, 80) ->
 40;
yeccgoto(expr_300, 83) ->
 40;
yeccgoto(expr_300, 88) ->
 40;
yeccgoto(expr_300, 93) ->
 40;
yeccgoto(expr_300, 97) ->
 40;
yeccgoto(expr_300, 102) ->
 40;
yeccgoto(expr_300, 104) ->
 40;
yeccgoto(expr_300, 111) ->
 40;
yeccgoto(expr_300, 125) ->
 40;
yeccgoto(expr_300, 137) ->
 40;
yeccgoto(expr_300, 142) ->
 40;
yeccgoto(expr_300, 167) ->
 170;
yeccgoto(expr_300, 180) ->
 181;
yeccgoto(expr_300, 182) ->
 40;
yeccgoto(expr_300, 184) ->
 40;
yeccgoto(expr_300, 186) ->
 40;
yeccgoto(expr_300, 187) ->
 40;
yeccgoto(expr_300, 190) ->
 40;
yeccgoto(expr_300, 194) ->
 40;
yeccgoto(expr_300, 201) ->
 40;
yeccgoto(expr_300, 204) ->
 40;
yeccgoto(expr_300, 239) ->
 40;
yeccgoto(expr_300, 257) ->
 40;
yeccgoto(expr_400, 14) ->
 41;
yeccgoto(expr_400, 18) ->
 41;
yeccgoto(expr_400, 20) ->
 41;
yeccgoto(expr_400, 25) ->
 41;
yeccgoto(expr_400, 28) ->
 41;
yeccgoto(expr_400, 31) ->
 41;
yeccgoto(expr_400, 33) ->
 41;
yeccgoto(expr_400, 54) ->
 41;
yeccgoto(expr_400, 63) ->
 41;
yeccgoto(expr_400, 70) ->
 41;
yeccgoto(expr_400, 75) ->
 41;
yeccgoto(expr_400, 80) ->
 41;
yeccgoto(expr_400, 83) ->
 41;
yeccgoto(expr_400, 88) ->
 41;
yeccgoto(expr_400, 93) ->
 41;
yeccgoto(expr_400, 97) ->
 41;
yeccgoto(expr_400, 102) ->
 41;
yeccgoto(expr_400, 104) ->
 41;
yeccgoto(expr_400, 111) ->
 41;
yeccgoto(expr_400, 125) ->
 41;
yeccgoto(expr_400, 137) ->
 41;
yeccgoto(expr_400, 142) ->
 41;
yeccgoto(expr_400, 167) ->
 41;
yeccgoto(expr_400, 180) ->
 41;
yeccgoto(expr_400, 182) ->
 41;
yeccgoto(expr_400, 184) ->
 41;
yeccgoto(expr_400, 186) ->
 41;
yeccgoto(expr_400, 187) ->
 41;
yeccgoto(expr_400, 190) ->
 41;
yeccgoto(expr_400, 194) ->
 41;
yeccgoto(expr_400, 201) ->
 41;
yeccgoto(expr_400, 204) ->
 41;
yeccgoto(expr_400, 239) ->
 41;
yeccgoto(expr_400, 257) ->
 41;
yeccgoto(expr_500, 14) ->
 42;
yeccgoto(expr_500, 18) ->
 42;
yeccgoto(expr_500, 20) ->
 42;
yeccgoto(expr_500, 25) ->
 42;
yeccgoto(expr_500, 28) ->
 42;
yeccgoto(expr_500, 31) ->
 42;
yeccgoto(expr_500, 33) ->
 42;
yeccgoto(expr_500, 54) ->
 42;
yeccgoto(expr_500, 63) ->
 42;
yeccgoto(expr_500, 70) ->
 42;
yeccgoto(expr_500, 75) ->
 42;
yeccgoto(expr_500, 80) ->
 42;
yeccgoto(expr_500, 83) ->
 42;
yeccgoto(expr_500, 88) ->
 42;
yeccgoto(expr_500, 93) ->
 42;
yeccgoto(expr_500, 97) ->
 42;
yeccgoto(expr_500, 102) ->
 42;
yeccgoto(expr_500, 104) ->
 42;
yeccgoto(expr_500, 111) ->
 42;
yeccgoto(expr_500, 125) ->
 42;
yeccgoto(expr_500, 137) ->
 42;
yeccgoto(expr_500, 142) ->
 42;
yeccgoto(expr_500, 162) ->
 171;
yeccgoto(expr_500, 167) ->
 42;
yeccgoto(expr_500, 180) ->
 42;
yeccgoto(expr_500, 182) ->
 42;
yeccgoto(expr_500, 184) ->
 42;
yeccgoto(expr_500, 186) ->
 42;
yeccgoto(expr_500, 187) ->
 42;
yeccgoto(expr_500, 190) ->
 42;
yeccgoto(expr_500, 194) ->
 42;
yeccgoto(expr_500, 201) ->
 42;
yeccgoto(expr_500, 204) ->
 42;
yeccgoto(expr_500, 239) ->
 42;
yeccgoto(expr_500, 257) ->
 42;
yeccgoto(expr_600, 14) ->
 43;
yeccgoto(expr_600, 18) ->
 43;
yeccgoto(expr_600, 20) ->
 43;
yeccgoto(expr_600, 25) ->
 43;
yeccgoto(expr_600, 28) ->
 43;
yeccgoto(expr_600, 31) ->
 43;
yeccgoto(expr_600, 33) ->
 43;
yeccgoto(expr_600, 54) ->
 43;
yeccgoto(expr_600, 63) ->
 43;
yeccgoto(expr_600, 70) ->
 43;
yeccgoto(expr_600, 75) ->
 43;
yeccgoto(expr_600, 80) ->
 43;
yeccgoto(expr_600, 83) ->
 43;
yeccgoto(expr_600, 88) ->
 43;
yeccgoto(expr_600, 93) ->
 43;
yeccgoto(expr_600, 97) ->
 43;
yeccgoto(expr_600, 102) ->
 43;
yeccgoto(expr_600, 104) ->
 43;
yeccgoto(expr_600, 111) ->
 43;
yeccgoto(expr_600, 125) ->
 43;
yeccgoto(expr_600, 137) ->
 43;
yeccgoto(expr_600, 142) ->
 43;
yeccgoto(expr_600, 155) ->
 157;
yeccgoto(expr_600, 162) ->
 43;
yeccgoto(expr_600, 167) ->
 43;
yeccgoto(expr_600, 180) ->
 43;
yeccgoto(expr_600, 182) ->
 43;
yeccgoto(expr_600, 184) ->
 43;
yeccgoto(expr_600, 186) ->
 43;
yeccgoto(expr_600, 187) ->
 43;
yeccgoto(expr_600, 190) ->
 43;
yeccgoto(expr_600, 194) ->
 43;
yeccgoto(expr_600, 201) ->
 43;
yeccgoto(expr_600, 204) ->
 43;
yeccgoto(expr_600, 239) ->
 43;
yeccgoto(expr_600, 257) ->
 43;
yeccgoto(expr_700, 14) ->
 44;
yeccgoto(expr_700, 18) ->
 44;
yeccgoto(expr_700, 20) ->
 44;
yeccgoto(expr_700, 25) ->
 44;
yeccgoto(expr_700, 28) ->
 44;
yeccgoto(expr_700, 31) ->
 44;
yeccgoto(expr_700, 33) ->
 44;
yeccgoto(expr_700, 54) ->
 44;
yeccgoto(expr_700, 60) ->
 106;
yeccgoto(expr_700, 63) ->
 44;
yeccgoto(expr_700, 70) ->
 44;
yeccgoto(expr_700, 75) ->
 44;
yeccgoto(expr_700, 80) ->
 44;
yeccgoto(expr_700, 83) ->
 44;
yeccgoto(expr_700, 88) ->
 44;
yeccgoto(expr_700, 93) ->
 44;
yeccgoto(expr_700, 97) ->
 44;
yeccgoto(expr_700, 102) ->
 44;
yeccgoto(expr_700, 104) ->
 44;
yeccgoto(expr_700, 111) ->
 44;
yeccgoto(expr_700, 125) ->
 44;
yeccgoto(expr_700, 137) ->
 44;
yeccgoto(expr_700, 142) ->
 44;
yeccgoto(expr_700, 155) ->
 44;
yeccgoto(expr_700, 162) ->
 44;
yeccgoto(expr_700, 167) ->
 44;
yeccgoto(expr_700, 180) ->
 44;
yeccgoto(expr_700, 182) ->
 44;
yeccgoto(expr_700, 184) ->
 44;
yeccgoto(expr_700, 186) ->
 44;
yeccgoto(expr_700, 187) ->
 44;
yeccgoto(expr_700, 190) ->
 44;
yeccgoto(expr_700, 194) ->
 44;
yeccgoto(expr_700, 201) ->
 44;
yeccgoto(expr_700, 204) ->
 44;
yeccgoto(expr_700, 239) ->
 44;
yeccgoto(expr_700, 257) ->
 44;
yeccgoto(expr_800, 14) ->
 45;
yeccgoto(expr_800, 18) ->
 45;
yeccgoto(expr_800, 20) ->
 45;
yeccgoto(expr_800, 25) ->
 45;
yeccgoto(expr_800, 28) ->
 45;
yeccgoto(expr_800, 31) ->
 45;
yeccgoto(expr_800, 33) ->
 45;
yeccgoto(expr_800, 54) ->
 45;
yeccgoto(expr_800, 60) ->
 45;
yeccgoto(expr_800, 63) ->
 45;
yeccgoto(expr_800, 70) ->
 45;
yeccgoto(expr_800, 75) ->
 45;
yeccgoto(expr_800, 80) ->
 45;
yeccgoto(expr_800, 83) ->
 45;
yeccgoto(expr_800, 88) ->
 45;
yeccgoto(expr_800, 93) ->
 45;
yeccgoto(expr_800, 97) ->
 45;
yeccgoto(expr_800, 102) ->
 45;
yeccgoto(expr_800, 104) ->
 45;
yeccgoto(expr_800, 111) ->
 45;
yeccgoto(expr_800, 125) ->
 45;
yeccgoto(expr_800, 137) ->
 45;
yeccgoto(expr_800, 142) ->
 45;
yeccgoto(expr_800, 155) ->
 45;
yeccgoto(expr_800, 162) ->
 45;
yeccgoto(expr_800, 167) ->
 45;
yeccgoto(expr_800, 180) ->
 45;
yeccgoto(expr_800, 182) ->
 45;
yeccgoto(expr_800, 184) ->
 45;
yeccgoto(expr_800, 186) ->
 45;
yeccgoto(expr_800, 187) ->
 45;
yeccgoto(expr_800, 190) ->
 45;
yeccgoto(expr_800, 194) ->
 45;
yeccgoto(expr_800, 201) ->
 45;
yeccgoto(expr_800, 204) ->
 45;
yeccgoto(expr_800, 239) ->
 45;
yeccgoto(expr_800, 257) ->
 45;
yeccgoto(expr_900, 14) ->
 46;
yeccgoto(expr_900, 18) ->
 46;
yeccgoto(expr_900, 20) ->
 46;
yeccgoto(expr_900, 25) ->
 46;
yeccgoto(expr_900, 28) ->
 46;
yeccgoto(expr_900, 31) ->
 46;
yeccgoto(expr_900, 33) ->
 46;
yeccgoto(expr_900, 54) ->
 46;
yeccgoto(expr_900, 60) ->
 46;
yeccgoto(expr_900, 63) ->
 46;
yeccgoto(expr_900, 70) ->
 46;
yeccgoto(expr_900, 75) ->
 46;
yeccgoto(expr_900, 80) ->
 46;
yeccgoto(expr_900, 83) ->
 46;
yeccgoto(expr_900, 88) ->
 46;
yeccgoto(expr_900, 93) ->
 46;
yeccgoto(expr_900, 97) ->
 46;
yeccgoto(expr_900, 102) ->
 46;
yeccgoto(expr_900, 104) ->
 46;
yeccgoto(expr_900, 111) ->
 46;
yeccgoto(expr_900, 125) ->
 46;
yeccgoto(expr_900, 137) ->
 46;
yeccgoto(expr_900, 142) ->
 46;
yeccgoto(expr_900, 155) ->
 46;
yeccgoto(expr_900, 162) ->
 46;
yeccgoto(expr_900, 167) ->
 46;
yeccgoto(expr_900, 180) ->
 46;
yeccgoto(expr_900, 182) ->
 46;
yeccgoto(expr_900, 184) ->
 46;
yeccgoto(expr_900, 186) ->
 46;
yeccgoto(expr_900, 187) ->
 46;
yeccgoto(expr_900, 190) ->
 46;
yeccgoto(expr_900, 194) ->
 46;
yeccgoto(expr_900, 201) ->
 46;
yeccgoto(expr_900, 204) ->
 46;
yeccgoto(expr_900, 239) ->
 46;
yeccgoto(expr_900, 257) ->
 46;
yeccgoto(expr_max, 14) ->
 47;
yeccgoto(expr_max, 18) ->
 47;
yeccgoto(expr_max, 20) ->
 47;
yeccgoto(expr_max, 24) ->
 213;
yeccgoto(expr_max, 25) ->
 47;
yeccgoto(expr_max, 28) ->
 47;
yeccgoto(expr_max, 31) ->
 47;
yeccgoto(expr_max, 33) ->
 47;
yeccgoto(expr_max, 54) ->
 47;
yeccgoto(expr_max, 60) ->
 47;
yeccgoto(expr_max, 63) ->
 47;
yeccgoto(expr_max, 70) ->
 47;
yeccgoto(expr_max, 75) ->
 47;
yeccgoto(expr_max, 80) ->
 47;
yeccgoto(expr_max, 83) ->
 47;
yeccgoto(expr_max, 88) ->
 47;
yeccgoto(expr_max, 93) ->
 47;
yeccgoto(expr_max, 97) ->
 47;
yeccgoto(expr_max, 102) ->
 47;
yeccgoto(expr_max, 104) ->
 47;
yeccgoto(expr_max, 111) ->
 47;
yeccgoto(expr_max, 125) ->
 47;
yeccgoto(expr_max, 137) ->
 47;
yeccgoto(expr_max, 142) ->
 47;
yeccgoto(expr_max, 146) ->
 147;
yeccgoto(expr_max, 155) ->
 47;
yeccgoto(expr_max, 162) ->
 47;
yeccgoto(expr_max, 167) ->
 47;
yeccgoto(expr_max, 180) ->
 47;
yeccgoto(expr_max, 182) ->
 47;
yeccgoto(expr_max, 184) ->
 47;
yeccgoto(expr_max, 186) ->
 47;
yeccgoto(expr_max, 187) ->
 47;
yeccgoto(expr_max, 190) ->
 47;
yeccgoto(expr_max, 194) ->
 47;
yeccgoto(expr_max, 201) ->
 47;
yeccgoto(expr_max, 204) ->
 47;
yeccgoto(expr_max, 214) ->
 215;
yeccgoto(expr_max, 216) ->
 228;
yeccgoto(expr_max, 230) ->
 213;
yeccgoto(expr_max, 239) ->
 47;
yeccgoto(expr_max, 257) ->
 47;
yeccgoto(exprs, 14) ->
 243;
yeccgoto(exprs, 18) ->
 48;
yeccgoto(exprs, 28) ->
 197;
yeccgoto(exprs, 54) ->
 48;
yeccgoto(exprs, 70) ->
 71;
yeccgoto(exprs, 80) ->
 82;
yeccgoto(exprs, 111) ->
 48;
yeccgoto(exprs, 125) ->
 48;
yeccgoto(exprs, 190) ->
 191;
yeccgoto(exprs, 257) ->
 259;
yeccgoto(form, 0) ->
 4;
yeccgoto(fun_clause, 50) ->
 116;
yeccgoto(fun_clause, 119) ->
 116;
yeccgoto(fun_clauses, 50) ->
 117;
yeccgoto(fun_clauses, 119) ->
 120;
yeccgoto(fun_expr, 14) ->
 51;
yeccgoto(fun_expr, 18) ->
 51;
yeccgoto(fun_expr, 20) ->
 51;
yeccgoto(fun_expr, 24) ->
 51;
yeccgoto(fun_expr, 25) ->
 51;
yeccgoto(fun_expr, 28) ->
 51;
yeccgoto(fun_expr, 31) ->
 51;
yeccgoto(fun_expr, 33) ->
 51;
yeccgoto(fun_expr, 54) ->
 51;
yeccgoto(fun_expr, 60) ->
 51;
yeccgoto(fun_expr, 63) ->
 51;
yeccgoto(fun_expr, 70) ->
 51;
yeccgoto(fun_expr, 75) ->
 51;
yeccgoto(fun_expr, 80) ->
 51;
yeccgoto(fun_expr, 83) ->
 51;
yeccgoto(fun_expr, 88) ->
 51;
yeccgoto(fun_expr, 93) ->
 51;
yeccgoto(fun_expr, 97) ->
 51;
yeccgoto(fun_expr, 102) ->
 51;
yeccgoto(fun_expr, 104) ->
 51;
yeccgoto(fun_expr, 111) ->
 51;
yeccgoto(fun_expr, 125) ->
 51;
yeccgoto(fun_expr, 137) ->
 51;
yeccgoto(fun_expr, 142) ->
 51;
yeccgoto(fun_expr, 146) ->
 51;
yeccgoto(fun_expr, 155) ->
 51;
yeccgoto(fun_expr, 162) ->
 51;
yeccgoto(fun_expr, 167) ->
 51;
yeccgoto(fun_expr, 180) ->
 51;
yeccgoto(fun_expr, 182) ->
 51;
yeccgoto(fun_expr, 184) ->
 51;
yeccgoto(fun_expr, 186) ->
 51;
yeccgoto(fun_expr, 187) ->
 51;
yeccgoto(fun_expr, 190) ->
 51;
yeccgoto(fun_expr, 194) ->
 51;
yeccgoto(fun_expr, 201) ->
 51;
yeccgoto(fun_expr, 204) ->
 51;
yeccgoto(fun_expr, 214) ->
 51;
yeccgoto(fun_expr, 216) ->
 51;
yeccgoto(fun_expr, 230) ->
 51;
yeccgoto(fun_expr, 239) ->
 51;
yeccgoto(fun_expr, 257) ->
 51;
yeccgoto(function, 0) ->
 5;
yeccgoto(function_call, 14) ->
 52;
yeccgoto(function_call, 18) ->
 52;
yeccgoto(function_call, 20) ->
 52;
yeccgoto(function_call, 25) ->
 52;
yeccgoto(function_call, 28) ->
 52;
yeccgoto(function_call, 31) ->
 52;
yeccgoto(function_call, 33) ->
 52;
yeccgoto(function_call, 54) ->
 52;
yeccgoto(function_call, 60) ->
 52;
yeccgoto(function_call, 63) ->
 52;
yeccgoto(function_call, 70) ->
 52;
yeccgoto(function_call, 75) ->
 52;
yeccgoto(function_call, 80) ->
 52;
yeccgoto(function_call, 83) ->
 52;
yeccgoto(function_call, 88) ->
 52;
yeccgoto(function_call, 93) ->
 52;
yeccgoto(function_call, 97) ->
 52;
yeccgoto(function_call, 102) ->
 52;
yeccgoto(function_call, 104) ->
 52;
yeccgoto(function_call, 111) ->
 52;
yeccgoto(function_call, 125) ->
 52;
yeccgoto(function_call, 137) ->
 52;
yeccgoto(function_call, 142) ->
 52;
yeccgoto(function_call, 155) ->
 52;
yeccgoto(function_call, 162) ->
 52;
yeccgoto(function_call, 167) ->
 52;
yeccgoto(function_call, 180) ->
 52;
yeccgoto(function_call, 182) ->
 52;
yeccgoto(function_call, 184) ->
 52;
yeccgoto(function_call, 186) ->
 52;
yeccgoto(function_call, 187) ->
 52;
yeccgoto(function_call, 190) ->
 52;
yeccgoto(function_call, 194) ->
 52;
yeccgoto(function_call, 201) ->
 52;
yeccgoto(function_call, 204) ->
 52;
yeccgoto(function_call, 239) ->
 52;
yeccgoto(function_call, 257) ->
 52;
yeccgoto(function_clause, 0) ->
 6;
yeccgoto(function_clause, 246) ->
 6;
yeccgoto(function_clauses, 0) ->
 7;
yeccgoto(function_clauses, 246) ->
 248;
yeccgoto(guard, 18) ->
 53;
yeccgoto(guard, 54) ->
 107;
yeccgoto(guard, 111) ->
 107;
yeccgoto(guard, 125) ->
 126;
yeccgoto(if_clause, 54) ->
 108;
yeccgoto(if_clause, 111) ->
 108;
yeccgoto(if_clauses, 54) ->
 109;
yeccgoto(if_clauses, 111) ->
 112;
yeccgoto(if_expr, 14) ->
 55;
yeccgoto(if_expr, 18) ->
 55;
yeccgoto(if_expr, 20) ->
 55;
yeccgoto(if_expr, 24) ->
 55;
yeccgoto(if_expr, 25) ->
 55;
yeccgoto(if_expr, 28) ->
 55;
yeccgoto(if_expr, 31) ->
 55;
yeccgoto(if_expr, 33) ->
 55;
yeccgoto(if_expr, 54) ->
 55;
yeccgoto(if_expr, 60) ->
 55;
yeccgoto(if_expr, 63) ->
 55;
yeccgoto(if_expr, 70) ->
 55;
yeccgoto(if_expr, 75) ->
 55;
yeccgoto(if_expr, 80) ->
 55;
yeccgoto(if_expr, 83) ->
 55;
yeccgoto(if_expr, 88) ->
 55;
yeccgoto(if_expr, 93) ->
 55;
yeccgoto(if_expr, 97) ->
 55;
yeccgoto(if_expr, 102) ->
 55;
yeccgoto(if_expr, 104) ->
 55;
yeccgoto(if_expr, 111) ->
 55;
yeccgoto(if_expr, 125) ->
 55;
yeccgoto(if_expr, 137) ->
 55;
yeccgoto(if_expr, 142) ->
 55;
yeccgoto(if_expr, 146) ->
 55;
yeccgoto(if_expr, 155) ->
 55;
yeccgoto(if_expr, 162) ->
 55;
yeccgoto(if_expr, 167) ->
 55;
yeccgoto(if_expr, 180) ->
 55;
yeccgoto(if_expr, 182) ->
 55;
yeccgoto(if_expr, 184) ->
 55;
yeccgoto(if_expr, 186) ->
 55;
yeccgoto(if_expr, 187) ->
 55;
yeccgoto(if_expr, 190) ->
 55;
yeccgoto(if_expr, 194) ->
 55;
yeccgoto(if_expr, 201) ->
 55;
yeccgoto(if_expr, 204) ->
 55;
yeccgoto(if_expr, 214) ->
 55;
yeccgoto(if_expr, 216) ->
 55;
yeccgoto(if_expr, 230) ->
 55;
yeccgoto(if_expr, 239) ->
 55;
yeccgoto(if_expr, 257) ->
 55;
yeccgoto(lc_expr, 97) ->
 99;
yeccgoto(lc_expr, 102) ->
 99;
yeccgoto(lc_expr, 239) ->
 99;
yeccgoto(lc_exprs, 97) ->
 100;
yeccgoto(lc_exprs, 102) ->
 103;
yeccgoto(lc_exprs, 239) ->
 241;
yeccgoto(list, 14) ->
 57;
yeccgoto(list, 18) ->
 57;
yeccgoto(list, 20) ->
 57;
yeccgoto(list, 24) ->
 57;
yeccgoto(list, 25) ->
 57;
yeccgoto(list, 28) ->
 57;
yeccgoto(list, 31) ->
 57;
yeccgoto(list, 33) ->
 57;
yeccgoto(list, 54) ->
 57;
yeccgoto(list, 60) ->
 57;
yeccgoto(list, 63) ->
 57;
yeccgoto(list, 70) ->
 57;
yeccgoto(list, 75) ->
 57;
yeccgoto(list, 80) ->
 57;
yeccgoto(list, 83) ->
 57;
yeccgoto(list, 88) ->
 57;
yeccgoto(list, 93) ->
 57;
yeccgoto(list, 97) ->
 57;
yeccgoto(list, 102) ->
 57;
yeccgoto(list, 104) ->
 57;
yeccgoto(list, 111) ->
 57;
yeccgoto(list, 125) ->
 57;
yeccgoto(list, 137) ->
 57;
yeccgoto(list, 142) ->
 57;
yeccgoto(list, 146) ->
 57;
yeccgoto(list, 155) ->
 57;
yeccgoto(list, 162) ->
 57;
yeccgoto(list, 167) ->
 57;
yeccgoto(list, 180) ->
 57;
yeccgoto(list, 182) ->
 57;
yeccgoto(list, 184) ->
 57;
yeccgoto(list, 186) ->
 57;
yeccgoto(list, 187) ->
 57;
yeccgoto(list, 190) ->
 57;
yeccgoto(list, 194) ->
 57;
yeccgoto(list, 201) ->
 57;
yeccgoto(list, 204) ->
 57;
yeccgoto(list, 214) ->
 57;
yeccgoto(list, 216) ->
 57;
yeccgoto(list, 230) ->
 57;
yeccgoto(list, 239) ->
 57;
yeccgoto(list, 257) ->
 57;
yeccgoto(list_comprehension, 14) ->
 58;
yeccgoto(list_comprehension, 18) ->
 58;
yeccgoto(list_comprehension, 20) ->
 58;
yeccgoto(list_comprehension, 24) ->
 58;
yeccgoto(list_comprehension, 25) ->
 58;
yeccgoto(list_comprehension, 28) ->
 58;
yeccgoto(list_comprehension, 31) ->
 58;
yeccgoto(list_comprehension, 33) ->
 58;
yeccgoto(list_comprehension, 54) ->
 58;
yeccgoto(list_comprehension, 60) ->
 58;
yeccgoto(list_comprehension, 61) ->
 94;
yeccgoto(list_comprehension, 63) ->
 58;
yeccgoto(list_comprehension, 70) ->
 58;
yeccgoto(list_comprehension, 75) ->
 58;
yeccgoto(list_comprehension, 80) ->
 58;
yeccgoto(list_comprehension, 83) ->
 58;
yeccgoto(list_comprehension, 88) ->
 58;
yeccgoto(list_comprehension, 93) ->
 58;
yeccgoto(list_comprehension, 97) ->
 58;
yeccgoto(list_comprehension, 102) ->
 58;
yeccgoto(list_comprehension, 104) ->
 58;
yeccgoto(list_comprehension, 111) ->
 58;
yeccgoto(list_comprehension, 125) ->
 58;
yeccgoto(list_comprehension, 137) ->
 58;
yeccgoto(list_comprehension, 142) ->
 58;
yeccgoto(list_comprehension, 146) ->
 58;
yeccgoto(list_comprehension, 155) ->
 58;
yeccgoto(list_comprehension, 162) ->
 58;
yeccgoto(list_comprehension, 167) ->
 58;
yeccgoto(list_comprehension, 180) ->
 58;
yeccgoto(list_comprehension, 182) ->
 58;
yeccgoto(list_comprehension, 184) ->
 58;
yeccgoto(list_comprehension, 186) ->
 58;
yeccgoto(list_comprehension, 187) ->
 58;
yeccgoto(list_comprehension, 190) ->
 58;
yeccgoto(list_comprehension, 194) ->
 58;
yeccgoto(list_comprehension, 201) ->
 58;
yeccgoto(list_comprehension, 204) ->
 58;
yeccgoto(list_comprehension, 214) ->
 58;
yeccgoto(list_comprehension, 216) ->
 58;
yeccgoto(list_comprehension, 230) ->
 58;
yeccgoto(list_comprehension, 239) ->
 58;
yeccgoto(list_comprehension, 257) ->
 58;
yeccgoto(list_op, 41) ->
 167;
yeccgoto(mult_op, 42) ->
 155;
yeccgoto(mult_op, 171) ->
 155;
yeccgoto(opt_bit_size_expr, 212) ->
 217;
yeccgoto(opt_bit_type_list, 217) ->
 219;
yeccgoto(prefix_op, 14) ->
 60;
yeccgoto(prefix_op, 18) ->
 60;
yeccgoto(prefix_op, 20) ->
 60;
yeccgoto(prefix_op, 24) ->
 214;
yeccgoto(prefix_op, 25) ->
 60;
yeccgoto(prefix_op, 28) ->
 60;
yeccgoto(prefix_op, 31) ->
 60;
yeccgoto(prefix_op, 33) ->
 60;
yeccgoto(prefix_op, 54) ->
 60;
yeccgoto(prefix_op, 63) ->
 60;
yeccgoto(prefix_op, 70) ->
 60;
yeccgoto(prefix_op, 75) ->
 60;
yeccgoto(prefix_op, 80) ->
 60;
yeccgoto(prefix_op, 83) ->
 60;
yeccgoto(prefix_op, 88) ->
 60;
yeccgoto(prefix_op, 93) ->
 60;
yeccgoto(prefix_op, 97) ->
 60;
yeccgoto(prefix_op, 102) ->
 60;
yeccgoto(prefix_op, 104) ->
 60;
yeccgoto(prefix_op, 111) ->
 60;
yeccgoto(prefix_op, 125) ->
 60;
yeccgoto(prefix_op, 137) ->
 60;
yeccgoto(prefix_op, 142) ->
 60;
yeccgoto(prefix_op, 155) ->
 60;
yeccgoto(prefix_op, 162) ->
 60;
yeccgoto(prefix_op, 167) ->
 60;
yeccgoto(prefix_op, 180) ->
 60;
yeccgoto(prefix_op, 182) ->
 60;
yeccgoto(prefix_op, 184) ->
 60;
yeccgoto(prefix_op, 186) ->
 60;
yeccgoto(prefix_op, 187) ->
 60;
yeccgoto(prefix_op, 190) ->
 60;
yeccgoto(prefix_op, 194) ->
 60;
yeccgoto(prefix_op, 201) ->
 60;
yeccgoto(prefix_op, 204) ->
 60;
yeccgoto(prefix_op, 230) ->
 214;
yeccgoto(prefix_op, 239) ->
 60;
yeccgoto(prefix_op, 257) ->
 60;
yeccgoto(query_expr, 14) ->
 62;
yeccgoto(query_expr, 18) ->
 62;
yeccgoto(query_expr, 20) ->
 62;
yeccgoto(query_expr, 24) ->
 62;
yeccgoto(query_expr, 25) ->
 62;
yeccgoto(query_expr, 28) ->
 62;
yeccgoto(query_expr, 31) ->
 62;
yeccgoto(query_expr, 33) ->
 62;
yeccgoto(query_expr, 54) ->
 62;
yeccgoto(query_expr, 60) ->
 62;
yeccgoto(query_expr, 63) ->
 62;
yeccgoto(query_expr, 70) ->
 62;
yeccgoto(query_expr, 75) ->
 62;
yeccgoto(query_expr, 80) ->
 62;
yeccgoto(query_expr, 83) ->
 62;
yeccgoto(query_expr, 88) ->
 62;
yeccgoto(query_expr, 93) ->
 62;
yeccgoto(query_expr, 97) ->
 62;
yeccgoto(query_expr, 102) ->
 62;
yeccgoto(query_expr, 104) ->
 62;
yeccgoto(query_expr, 111) ->
 62;
yeccgoto(query_expr, 125) ->
 62;
yeccgoto(query_expr, 137) ->
 62;
yeccgoto(query_expr, 142) ->
 62;
yeccgoto(query_expr, 146) ->
 62;
yeccgoto(query_expr, 155) ->
 62;
yeccgoto(query_expr, 162) ->
 62;
yeccgoto(query_expr, 167) ->
 62;
yeccgoto(query_expr, 180) ->
 62;
yeccgoto(query_expr, 182) ->
 62;
yeccgoto(query_expr, 184) ->
 62;
yeccgoto(query_expr, 186) ->
 62;
yeccgoto(query_expr, 187) ->
 62;
yeccgoto(query_expr, 190) ->
 62;
yeccgoto(query_expr, 194) ->
 62;
yeccgoto(query_expr, 201) ->
 62;
yeccgoto(query_expr, 204) ->
 62;
yeccgoto(query_expr, 214) ->
 62;
yeccgoto(query_expr, 216) ->
 62;
yeccgoto(query_expr, 230) ->
 62;
yeccgoto(query_expr, 239) ->
 62;
yeccgoto(query_expr, 257) ->
 62;
yeccgoto(receive_expr, 14) ->
 64;
yeccgoto(receive_expr, 18) ->
 64;
yeccgoto(receive_expr, 20) ->
 64;
yeccgoto(receive_expr, 24) ->
 64;
yeccgoto(receive_expr, 25) ->
 64;
yeccgoto(receive_expr, 28) ->
 64;
yeccgoto(receive_expr, 31) ->
 64;
yeccgoto(receive_expr, 33) ->
 64;
yeccgoto(receive_expr, 54) ->
 64;
yeccgoto(receive_expr, 60) ->
 64;
yeccgoto(receive_expr, 63) ->
 64;
yeccgoto(receive_expr, 70) ->
 64;
yeccgoto(receive_expr, 75) ->
 64;
yeccgoto(receive_expr, 80) ->
 64;
yeccgoto(receive_expr, 83) ->
 64;
yeccgoto(receive_expr, 88) ->
 64;
yeccgoto(receive_expr, 93) ->
 64;
yeccgoto(receive_expr, 97) ->
 64;
yeccgoto(receive_expr, 102) ->
 64;
yeccgoto(receive_expr, 104) ->
 64;
yeccgoto(receive_expr, 111) ->
 64;
yeccgoto(receive_expr, 125) ->
 64;
yeccgoto(receive_expr, 137) ->
 64;
yeccgoto(receive_expr, 142) ->
 64;
yeccgoto(receive_expr, 146) ->
 64;
yeccgoto(receive_expr, 155) ->
 64;
yeccgoto(receive_expr, 162) ->
 64;
yeccgoto(receive_expr, 167) ->
 64;
yeccgoto(receive_expr, 180) ->
 64;
yeccgoto(receive_expr, 182) ->
 64;
yeccgoto(receive_expr, 184) ->
 64;
yeccgoto(receive_expr, 186) ->
 64;
yeccgoto(receive_expr, 187) ->
 64;
yeccgoto(receive_expr, 190) ->
 64;
yeccgoto(receive_expr, 194) ->
 64;
yeccgoto(receive_expr, 201) ->
 64;
yeccgoto(receive_expr, 204) ->
 64;
yeccgoto(receive_expr, 214) ->
 64;
yeccgoto(receive_expr, 216) ->
 64;
yeccgoto(receive_expr, 230) ->
 64;
yeccgoto(receive_expr, 239) ->
 64;
yeccgoto(receive_expr, 257) ->
 64;
yeccgoto(record_expr, 14) ->
 65;
yeccgoto(record_expr, 18) ->
 65;
yeccgoto(record_expr, 20) ->
 65;
yeccgoto(record_expr, 25) ->
 65;
yeccgoto(record_expr, 28) ->
 65;
yeccgoto(record_expr, 31) ->
 65;
yeccgoto(record_expr, 33) ->
 65;
yeccgoto(record_expr, 54) ->
 65;
yeccgoto(record_expr, 60) ->
 65;
yeccgoto(record_expr, 63) ->
 65;
yeccgoto(record_expr, 70) ->
 65;
yeccgoto(record_expr, 75) ->
 65;
yeccgoto(record_expr, 80) ->
 65;
yeccgoto(record_expr, 83) ->
 65;
yeccgoto(record_expr, 88) ->
 65;
yeccgoto(record_expr, 93) ->
 65;
yeccgoto(record_expr, 97) ->
 65;
yeccgoto(record_expr, 102) ->
 65;
yeccgoto(record_expr, 104) ->
 65;
yeccgoto(record_expr, 111) ->
 65;
yeccgoto(record_expr, 125) ->
 65;
yeccgoto(record_expr, 137) ->
 65;
yeccgoto(record_expr, 142) ->
 65;
yeccgoto(record_expr, 155) ->
 65;
yeccgoto(record_expr, 162) ->
 65;
yeccgoto(record_expr, 167) ->
 65;
yeccgoto(record_expr, 180) ->
 65;
yeccgoto(record_expr, 182) ->
 65;
yeccgoto(record_expr, 184) ->
 65;
yeccgoto(record_expr, 186) ->
 65;
yeccgoto(record_expr, 187) ->
 65;
yeccgoto(record_expr, 190) ->
 65;
yeccgoto(record_expr, 194) ->
 65;
yeccgoto(record_expr, 201) ->
 65;
yeccgoto(record_expr, 204) ->
 65;
yeccgoto(record_expr, 239) ->
 65;
yeccgoto(record_expr, 257) ->
 65;
yeccgoto(record_field, 131) ->
 133;
yeccgoto(record_field, 140) ->
 133;
yeccgoto(record_fields, 131) ->
 134;
yeccgoto(record_fields, 140) ->
 141;
yeccgoto(record_tuple, 128) ->
 130;
yeccgoto(record_tuple, 235) ->
 237;
yeccgoto(rule, 0) ->
 8;
yeccgoto(rule_body, 17) ->
 240;
yeccgoto(rule_body, 255) ->
 240;
yeccgoto(rule_clause, 0) ->
 9;
yeccgoto(rule_clause, 11) ->
 9;
yeccgoto(rule_clauses, 0) ->
 10;
yeccgoto(rule_clauses, 11) ->
 13;
yeccgoto(strings, 14) ->
 67;
yeccgoto(strings, 18) ->
 67;
yeccgoto(strings, 20) ->
 67;
yeccgoto(strings, 24) ->
 67;
yeccgoto(strings, 25) ->
 67;
yeccgoto(strings, 28) ->
 67;
yeccgoto(strings, 31) ->
 67;
yeccgoto(strings, 33) ->
 67;
yeccgoto(strings, 54) ->
 67;
yeccgoto(strings, 60) ->
 67;
yeccgoto(strings, 63) ->
 67;
yeccgoto(strings, 66) ->
 74;
yeccgoto(strings, 70) ->
 67;
yeccgoto(strings, 75) ->
 67;
yeccgoto(strings, 80) ->
 67;
yeccgoto(strings, 83) ->
 67;
yeccgoto(strings, 88) ->
 67;
yeccgoto(strings, 93) ->
 67;
yeccgoto(strings, 97) ->
 67;
yeccgoto(strings, 102) ->
 67;
yeccgoto(strings, 104) ->
 67;
yeccgoto(strings, 111) ->
 67;
yeccgoto(strings, 125) ->
 67;
yeccgoto(strings, 137) ->
 67;
yeccgoto(strings, 142) ->
 67;
yeccgoto(strings, 146) ->
 67;
yeccgoto(strings, 155) ->
 67;
yeccgoto(strings, 162) ->
 67;
yeccgoto(strings, 167) ->
 67;
yeccgoto(strings, 180) ->
 67;
yeccgoto(strings, 182) ->
 67;
yeccgoto(strings, 184) ->
 67;
yeccgoto(strings, 186) ->
 67;
yeccgoto(strings, 187) ->
 67;
yeccgoto(strings, 190) ->
 67;
yeccgoto(strings, 194) ->
 67;
yeccgoto(strings, 201) ->
 67;
yeccgoto(strings, 204) ->
 67;
yeccgoto(strings, 214) ->
 67;
yeccgoto(strings, 216) ->
 67;
yeccgoto(strings, 230) ->
 67;
yeccgoto(strings, 239) ->
 67;
yeccgoto(strings, 257) ->
 67;
yeccgoto(tail, 200) ->
 203;
yeccgoto(tail, 207) ->
 208;
yeccgoto(tuple, 14) ->
 68;
yeccgoto(tuple, 18) ->
 68;
yeccgoto(tuple, 20) ->
 68;
yeccgoto(tuple, 24) ->
 68;
yeccgoto(tuple, 25) ->
 68;
yeccgoto(tuple, 28) ->
 68;
yeccgoto(tuple, 31) ->
 68;
yeccgoto(tuple, 33) ->
 68;
yeccgoto(tuple, 54) ->
 68;
yeccgoto(tuple, 60) ->
 68;
yeccgoto(tuple, 63) ->
 68;
yeccgoto(tuple, 70) ->
 68;
yeccgoto(tuple, 75) ->
 68;
yeccgoto(tuple, 80) ->
 68;
yeccgoto(tuple, 83) ->
 68;
yeccgoto(tuple, 88) ->
 68;
yeccgoto(tuple, 93) ->
 68;
yeccgoto(tuple, 97) ->
 68;
yeccgoto(tuple, 102) ->
 68;
yeccgoto(tuple, 104) ->
 68;
yeccgoto(tuple, 111) ->
 68;
yeccgoto(tuple, 125) ->
 68;
yeccgoto(tuple, 137) ->
 68;
yeccgoto(tuple, 142) ->
 68;
yeccgoto(tuple, 146) ->
 68;
yeccgoto(tuple, 155) ->
 68;
yeccgoto(tuple, 162) ->
 68;
yeccgoto(tuple, 167) ->
 68;
yeccgoto(tuple, 180) ->
 68;
yeccgoto(tuple, 182) ->
 68;
yeccgoto(tuple, 184) ->
 68;
yeccgoto(tuple, 186) ->
 68;
yeccgoto(tuple, 187) ->
 68;
yeccgoto(tuple, 190) ->
 68;
yeccgoto(tuple, 194) ->
 68;
yeccgoto(tuple, 201) ->
 68;
yeccgoto(tuple, 204) ->
 68;
yeccgoto(tuple, 214) ->
 68;
yeccgoto(tuple, 216) ->
 68;
yeccgoto(tuple, 230) ->
 68;
yeccgoto(tuple, 239) ->
 68;
yeccgoto(tuple, 257) ->
 68;
yeccgoto(__Symbol, __State) ->
 exit({__Symbol, __State, missing_in_goto_table}).


